--
-- OrganizedPanic
--
-- Copyright (c) 2021 Dana Basken
--

local _, NS = ...

local VERSION="9.1.0.8"

OrganizedPanic = LibStub("AceAddon-3.0"):NewAddon(NS.ADDON_NAME, "AceConsole-3.0", "AceTimer-3.0", "AceEvent-3.0", "AceComm-3.0")

print(OrganizedPanicCommon:PrefixString(NS.ADDON_NAME, ("v%s"):format(VERSION), NS.PREFIX_COLOR))

local OPTIONS = NS:BuildOptions(NS.ADDON_NAME, OrganizedPanic)

local AceGUI = LibStub("AceGUI-3.0") -- needed for AceConfigDialog
local AceEvent = LibStub("AceEvent-3.0")
local AceTimer = LibStub("AceTimer-3.0")

local COOLDOWN_UPDATE_INTERVAL = 5
local UPDATE_SPELL_COOLDOWNS_INTERVAL = 0.5
local SEND_VERSION_MESSAGE_INTERVAL = 300
local INSTANT_SPELL_AUTO_STOP_TIMEOUT = 0.5

function OrganizedPanic:OnInitialize()
  self.db = LibStub("AceDB-3.0"):New("OrganizedPanicDB", NS.DEFAULTS, true)

  NS.db = self.db
  NS.debug = OrganizedPanic.debug
  NS.trace = OrganizedPanic.trace
  NS.ObjectToString = OrganizedPanicCommon.ObjectToString
  NS.VERSION = VERSION

  self.minimapIcon = OrganizedPanicMinimapIcon:Create(("%sMinimap"):format(NS.ADDON_NAME), "Interface/AddOns/OrganizedPanic/images/op_icon_normal.tga", self.db, function() self.window:Toggle() end)

  LibStub("AceConfig-3.0"):RegisterOptionsTable(NS.ADDON_NAME, OPTIONS)
  LibStub("AceConfigDialog-3.0"):AddToBlizOptions(NS.ADDON_NAME)

  self.flasher = BorderFlasher:Create("Interface\\Addons\\OrganizedPanic\\images\\border.tga", self.db.profile.borderColor)
  self.castingStates = {}
  self.latestVersionSeen = VERSION
  self.members = OrganizedPanicMembers:Create()
  self.window = OrganizedPanicWindow:Create(self.members)

  OrganizedPanic:RegisterEvent("UNIT_SPELLCAST_SUCCEEDED", "SPELLCAST_EVENT_HANDLER")
  OrganizedPanic:RegisterEvent("UNIT_SPELLCAST_START", "SPELLCAST_EVENT_HANDLER")
  OrganizedPanic:RegisterEvent("UNIT_SPELLCAST_STOP", "SPELLCAST_EVENT_HANDLER")
  OrganizedPanic:RegisterEvent("UNIT_SPELLCAST_CHANNEL_START", "SPELLCAST_EVENT_HANDLER")
  OrganizedPanic:RegisterEvent("UNIT_SPELLCAST_CHANNEL_STOP", "SPELLCAST_EVENT_HANDLER")

  OrganizedPanic:RegisterEvent("GROUP_ROSTER_UPDATE", "GROUP_UPDATE_HANDLER")
  OrganizedPanic:RegisterEvent("GROUP_JOINED", "GROUP_UPDATE_HANDLER")
  OrganizedPanic:RegisterEvent("GROUP_LEFT", "GROUP_UPDATE_HANDLER")
  OrganizedPanic:RegisterEvent("PLAYER_SPECIALIZATION_CHANGED")

  AceEvent.RegisterMessage(self, "CASTING_STATE_CHANGE")
  AceEvent.RegisterMessage(self, "SPELL_STATE_CHANGE")
  AceEvent.RegisterMessage(self, "SPELL_COOLDOWN")
  AceEvent.RegisterMessage(self, "MEMBER_INSPECTED")

  OrganizedPanic:RegisterComm(NS.ADDON_NAME)
  OrganizedPanic:RegisterChatCommand(NS.CHAT_COMMAND_NAME, "HandleChatCommand")
end

function OrganizedPanic:OnEnable()
  AceTimer.ScheduleRepeatingTimer(self, "UpdateSpellCooldowns", UPDATE_SPELL_COOLDOWNS_INTERVAL)
  AceTimer.ScheduleRepeatingTimer(self, "SendVersionMessages", SEND_VERSION_MESSAGE_INTERVAL)
  self:GROUP_UPDATE_HANDLER()
end

function OrganizedPanic:OnDisable()
  AceTimer:CancelAllTimers()
end

function OrganizedPanic:PLAYER_SPECIALIZATION_CHANGED(event, unitId)
  local member = self.members:FindMemberByGroupName(unitId)
  NS:debug(event, unitId, member)
  if member then member:SetNeedsInspect() end
  self.members:Update()
  self:SendGroupAddonMessage("spec_change")
end

function OrganizedPanic:GROUP_UPDATE_HANDLER(...)
  NS:debug("GROUP_UPDATE_HANDLER", ...)
  self.members:Update()
end

function OrganizedPanic:MEMBER_INSPECTED(event, name, unitName)
  self:SendGroupAddonMessage("send_spells")
end

function OrganizedPanic:CASTING_STATE_CHANGE(event, name, state, spellId)
  NS:trace(event, name, state, spellId)
  local member = self.members:FindMemberByName(name)
  if not member then return end
  if state == "SPELL_START" then
    local spell = member:FindMemberSpellById(spellId)
    if spell then member:SetSpellStart(spell, GetTime()) end -- set initial start, will get updated if member has addon
    if self.db.profile.includeRole then
      if UnitGroupRolesAssigned("player") == member:GetRole() then self:Panic(member:GetName(), spellId) end
    else
      self:Panic(member:GetName(), spellId)
    end
  end
end

function OrganizedPanic:SPELL_STATE_CHANGE(event, name, spellId, state)
  NS:trace(event, name, spellId, state)
  local member = self.members:FindMemberByName(name)
  if not member then return end
  if member:IsPlayer() then
    local spell = member:FindMemberSpellById(spellId)
    if spell then self:SendSpellMessage(member, spell) end
  end
end

function OrganizedPanic:SPELL_COOLDOWN(event, name, spellId, remaining)
  NS:trace(event, name, spellId, remaining)
  local member = self.members:FindMemberByName(name)
  if not member then return end
  if member:IsPlayer() then
    local now = GetTime()
    if not self.lastSpellCooldownSent then self.lastSpellCooldownSent = GetTime() end
    if now - self.lastSpellCooldownSent <= COOLDOWN_UPDATE_INTERVAL then return end
    local spell = member:FindMemberSpellById(spellId)
    if spell then
      self:SendSpellMessage(member, spell)
      self.lastSpellCooldownSent = now
    end
  end
end

-- every second
function OrganizedPanic:UpdateSpellCooldowns()
  if not OrganizedPanicCommon:InGlobalCooldown() then
    self:UpdatePlayerSpellCooldowns()
  end
  self.members:UpdateCooldownsRemaining()
end

-- every second, if not in GCD
function OrganizedPanic:UpdatePlayerSpellCooldowns()
  local member = self.members:FindPlayer()
  if not member then return end
  for _, spell in pairs(member:GetSpells()) do
    local start, duration, enabled, modRate = GetSpellCooldown(spell.id);
    member:SetSpellStart(spell, start, duration)
  end
end

function OrganizedPanic:WhisperMySpellStates(name)
  local member = self.members:FindPlayer()
  if not member then return end
  local hasSpells = false
  for _, spell in pairs(member.spells) do
    hasSpells = true
    self:SendSpellMessage(member, spell, name)
  end
  if not hasSpells then
    self:SendGroupAddonMessage("no_spells")
  end
end

function OrganizedPanic:SendSpellMessage(member, spell, name)
  local message = ("spell:%s:%s:%s:%d:%d:%d"):format(member.name, spell.id, spell.state, spell.start, spell.remaining, spell.duration)
  self:SendGroupAddonMessage(message)
end

function OrganizedPanic:SendGroupAddonMessage(message, priority)
  local channel = OrganizedPanicCommon:GetChannel()
  if not channel then return end
  self:SendCommMessageInternal(message, channel, nil, priority)
end

function OrganizedPanic:SendVersionMessages()
  local channel = OrganizedPanicCommon:GetChannel()
  if channel then
    self:SendVersionMessage(channel)
  else
    if IsInGuild() then self:SendVersionMessage("GUILD") end
  end
end

function OrganizedPanic:SendVersionMessage(channel)
  self:SendCommMessageInternal(("version:%s"):format(self.latestVersionSeen), channel)
end

function OrganizedPanic:SendCommMessageInternal(message, channel, name, priority)
  NS:debug("SendCommMessage", message, channel, name, priority)
  if name and name == OrganizedPanicCommon:GetPlayerNameWithRealm() then
    NS:debug("SendCommMessage", "not sending to myself")
    return
  end
  self:SendCommMessage(NS.ADDON_NAME, message, channel, name, priority)
end

-- handle all API spell events, and turn them into SPELL_START or SPELL_STOP, and have member handle the state change
function OrganizedPanic:SPELLCAST_EVENT_HANDLER(event, unit, guid, spellId)
  if not NS.db.profile.standalone and not OrganizedPanicCommon:InGroup() then return end
  -- find member
  local member = self.members:FindMemberByUnitName(unit)
  if not member then return end
  -- convert API event to simple start/stop event
  if event == "UNIT_SPELLCAST_SUCCEEDED" or event == "UNIT_SPELLCAST_START" or event == "UNIT_SPELLCAST_CHANNEL_START" then event = "SPELL_START" end
  if event == "UNIT_SPELLCAST_STOP" or event == "UNIT_SPELLCAST_CHANNEL_STOP" then event = "SPELL_STOP" end
  -- apply spell modifications
  if event == "SPELL_START" then member:ApplySpellModifications(spellId) end
   -- if this is a panic spell, update spell state
  if not NS.PANIC_SPELLS[spellId] then return end
  member:SetCastingState(event, spellId)
  if NS.PANIC_SPELLS[spellId].type == "INSTANT" then
    C_Timer.After(INSTANT_SPELL_AUTO_STOP_TIMEOUT, function() member:SetCastingState("SPELL_STOP", spellId) end)
  end
end

function OrganizedPanic:Panic(caster, id, force)
  if not force then
    if not self.db.profile.includeSelf and caster == OrganizedPanicCommon:GetPlayerNameWithRealm() then return end
  end
  self:PlayAlertSound()
  self:ShowRaidWarning(caster, id)
  self:FlashBorder()
end

function OrganizedPanic:PlayAlertSound()
  if self.db.profile.playSound then
    PlaySoundFile(NS.SOUNDS[self.db.profile.sound], "Master")
  end
end

function OrganizedPanic:ShowRaidWarning(caster, id)
  local spell = GetSpellInfo(id)
  if not spell then spell = id end
  local message = ("%s started casting %s"):format(caster, spell)
  RaidNotice_AddMessage(RaidWarningFrame, message, ChatTypeInfo["RAID_WARNING"])
end

function OrganizedPanic:FlashBorder()
  if self.db.profile.flashBorder then
    self.flasher:StartFlashing()
  end
end

function OrganizedPanic:GetOption(info)
  local property = info.option.arg
  return self.db.profile[property]
end

function OrganizedPanic:SetOption(info, value)
  local property = info.option.arg
  self.db.profile[property] = value
end

function OrganizedPanic:GetMimimapVisible(info)
  return not self.db.profile.minimap.hide
end

function OrganizedPanic:SetMimimapVisible(info, value)
  self.db.profile.minimap.hide = not value
  self.minimapIcon:UpdateVisibility()
end

function OrganizedPanic:SetBorderColor(info, r, g, b, a)
  self.db.profile.borderColor.r = r
  self.db.profile.borderColor.g = g
  self.db.profile.borderColor.b = b
  self.flasher:SetBorderColor(self.db.profile.borderColor)
  self.flasher:StartFlashing()
end

function OrganizedPanic:GetBorderColor(info)
  return self.db.profile.borderColor.r, self.db.profile.borderColor.g, self.db.profile.borderColor.b, 1.0
end

function OrganizedPanic:AddonPrint(format, ...)
  print(OrganizedPanicCommon:PrefixString(NS.ADDON_NAME, string.format(format, ...), NS.PREFIX_COLOR))
end

function OrganizedPanic:HandleChatCommand(input)
  if not input or input:trim() == "" then
    LibStub("AceConfigDialog-3.0"):Open("OrganizedPanic")
  else
    if input == "help" then
      self:AddonPrint("%s", "Commands:")
    elseif input == "options" then
      print(NS:ObjectToString(self.db.profile))
    elseif input == "panic" then
      self:Panic(OrganizedPanicCommon:GetPlayerNameWithRealm(), "Debug", true)
    elseif input == "settings" then
      self:ShowSettings()
    elseif input == "debug" then
      if self.db.profile.debug then self.db.profile.debug = false else self.db.profile.debug = true end
      self:ShowSettings()
    elseif input == "trace" then
      if self.db.profile.trace then self.db.profile.trace = false else self.db.profile.trace = true end
      self:ShowSettings()
    elseif input == "normal" then
      self.db.profile.debug = false
      self.db.profile.trace = false
      self.db.profile.standalone = false
      self.members:Update()
      self:ShowSettings()
    elseif input == "instance" then
      if self.db.profile.requireInstance then self.db.profile.requireInstance = false else self.db.profile.requireInstance = true end
    elseif input == "standalone" then
      if self.db.profile.standalone then self.db.profile.standalone = false else self.db.profile.standalone = true end
      self.members:Update()
      self:ShowSettings()
    elseif input == "members" then
      for _, member in pairs(self.members:GetMembers()) do
        print(member.name, member:GetClass(), member:GetSpecialization(), member:GetRole(), member:GetGroupName())
        for _, spell in pairs(member:GetSpells()) do
          print("  ", spell.name, spell.id, spell.state, spell.start, spell.duration, spell.remaining)
        end
      end
    elseif input == "dump" then
      for _, member in pairs(self.members:GetMembers()) do
        print(NS:ObjectToString(member))
      end
    elseif input == "minimap" then
      self.minimapIcon:Toggle()
    elseif input == "toggle" then
      self.window:Toggle()
    elseif input == "show" then
      self.window:Show()
    elseif input == "hide" then
      self.window:Hide()
    elseif input == "lock" then
      self.window:Lock()
    elseif input == "unlock" then
      self.window:Unlock()
    else
      self:AddonPrint("unknown command: '%s', try '/%s help'", input, NS.CHAT_COMMAND_NAME)
    end
  end
end

function OrganizedPanic:ShowSettings()
  self:AddonPrint("SETTINGS:")
  self:AddonPrint("  debug: %s", OrganizedPanicCommon:BooleanToString(self.db.profile.debug))
  self:AddonPrint("  trace: %s", OrganizedPanicCommon:BooleanToString(self.db.profile.trace))
  self:AddonPrint("  includeRole: %s", OrganizedPanicCommon:BooleanToString(self.db.profile.includeRole))
  self:AddonPrint("  includeSelf: %s", OrganizedPanicCommon:BooleanToString(self.db.profile.includeSelf))
  self:AddonPrint("  requireInstance: %s", OrganizedPanicCommon:BooleanToString(self.db.profile.requireInstance))
  self:AddonPrint("  standalone: %s", OrganizedPanicCommon:BooleanToString(self.db.profile.standalone))
  self:AddonPrint("  flashBorder: %s", OrganizedPanicCommon:BooleanToString(self.db.profile.flashBorder))
  self:AddonPrint("  sound: %s", self.db.profile.sound)
  self:AddonPrint("  playSound: %s", OrganizedPanicCommon:BooleanToString(self.db.profile.playSound))
end

function OrganizedPanic:HandleSpellMessage(data)
  local name, spellId, state, start, remaining, duration = strsplit(":", data)
  local member = self.members:FindMemberByName(name)
  if not member then
    NS:debug("HandleSpellMessage", "could not find member", name)
    return
  end
  local spell = member:FindMemberSpellById(tonumber(spellId))
  if not spell then
    NS:debug("HandleSpellMessage", "could not find spell", member.name, spellId)
    return
  end
  member:SetSpellStart(spell, tonumber(start), tonumber(duration), state)
end

function OrganizedPanic:HandleSpecChangeMessage(name)
  local member = self.members:FindMemberByName(name)
  if not member then
    NS:debug("HandleSpecChangeMessage", "could not find member", name)
    return
  end
  NS:debug("HandleSpecChangeMessage", member.name)
  member:SetNeedsInspect()
end

function OrganizedPanic:HandleVersionMessage(version)
  if OrganizedPanicCommon:IsNewerVersion(version, self.latestVersionSeen) then
    self:AddonPrint("new version available: %s", version)
    self.latestVersionSeen = version
  end
end

function OrganizedPanic:OnCommReceived(prefix, message, distribution, sender)
  local event, data = strsplit(":", message, 2)
  local sender = OrganizedPanicCommon:GetNameWithRealm(sender)
  NS:debug("OnCommReceived", event, data, sender, distribution)
  if sender ~= OrganizedPanicCommon:GetPlayerNameWithRealm() then
    if event == "version" then self:HandleVersionMessage(data) end
    if event == "send_spells" then self:WhisperMySpellStates(sender) end
    if event == "spell" then self:HandleSpellMessage(data) end
    if event == "spec_change" then self:HandleSpecChangeMessage(sender) end
    local member = self.members:FindMemberByName(sender)
    if member then member:SetHasAddon() end
  end
end

function OrganizedPanic:debug(where, ...)
  if NS.db.profile.debug then print(OrganizedPanicCommon:ColorString("op-debug:", "8080FF"), OrganizedPanicCommon:ColorString(where, "808080"), ...) end
end

function OrganizedPanic:trace(where, ...)
  if NS.db.profile.trace then print(OrganizedPanicCommon:ColorString("op-trace:", "FF80FF"), OrganizedPanicCommon:ColorString(where, "808080"), ...) end
end
