--
-- OrganizedPanicMembers
--
-- Copyright (c) 2021 Dana Basken
--

local _, NS = ...

local AceEvent = LibStub("AceEvent-3.0")
local AceTimer = LibStub("AceTimer-3.0")

OrganizedPanicMembers = {}
OrganizedPanicMembers.__index = OrganizedPanicMembers

local PERIODIC_INSPECTIONS_INTERVAL = 1

function OrganizedPanicMembers:Create()
  local instance = {}
  setmetatable(instance, OrganizedPanicMembers)
  instance:Construct()
  return instance
end

function OrganizedPanicMembers:Construct()
  self.members = {}
  AceEvent.RegisterEvent(self, "INSPECT_READY")
  self:Update()
  AceTimer.ScheduleRepeatingTimer(self, "PeriodicInspections", PERIODIC_INSPECTIONS_INTERVAL)
end

function OrganizedPanicMembers:INSPECT_READY(event, unitId)
  local member = self:FindMemberByUnitId(unitId)
  NS:debug(event, unitId, member and member.name, member and member.groupName, member and member.unitName, member and member.groupType)
  if not member then return end
  member:Update()
  local _, specialization, _, icon, role = GetSpecializationInfoByID(GetInspectSpecialization(member:GetGroupName()))
  self:UpdateMemberSpecialization(member, specialization, icon)
end

function OrganizedPanicMembers:PeriodicInspections()
  for _, member in pairs(self:GetMembers()) do
    if member:NeedsInspect() then
      NS:debug("PeriodicInspections", "needs inspection", member.name, member.groupName)
      self:InspectMember(member)
    end
  end
end

function OrganizedPanicMembers:Update()
  if NS.db.profile.standalone or OrganizedPanicCommon:InGroup("player") then
    self:UpdateGroup()
  else
    self.members = {}
  end
  AceEvent:SendMessage("MEMBERS_UPDATED")
end

-- every second
function OrganizedPanicMembers:UpdateCooldownsRemaining()
  for _, member in pairs(self:GetMembers()) do member:UpdateSpellCooldowns() end
end

function OrganizedPanicMembers:UpdateGroup()
  self:AddOrUpdateMember(OrganizedPanicCommon:GetPlayerNameWithRealm())
  local groupType = OrganizedPanicCommon:GetGroupType()
  if not groupType then return end
  local members = GetNumGroupMembers()
  for i = 1, members do
    local groupName = ("%s%d"):format(groupType, i)
    local name = OrganizedPanicCommon:GetUnitNameWithRealm(groupName)
    if name and UnitIsConnected(groupName) and self:GroupNameIsSameRole(groupName) then
        self:AddOrUpdateMember(name)
    else
      local member = self:FindMemberByName(name)
      if member then self:DeleteMember(member) end
    end
  end
  for _, member in pairs(self.members) do
    local groupName = member:GetGroupName()
    if groupName and not UnitIsConnected(groupName) or not self:GroupNameIsSameRole(groupName) then
      NS:debug("UpdateGroup", "member gone or not same role", member.name, member:GetRole())
    end
  end
end

function OrganizedPanicMembers:AddOrUpdateMember(name)
  NS:trace("AddOrUpdateMember", name)
  local member = self:FindMemberByName(name)
  if not member then member = self:AddMember(name) end
end

function OrganizedPanicMembers:GroupNameIsSameRole(groupName)
  if NS.db.profile.includeRole then
    local player = self:FindPlayer()
    if player and player:GetRole() ~= UnitGroupRolesAssigned(groupName) then return false end
  end
  return true
end

function OrganizedPanicMembers:AddMember(name)
  local member = OrganizedPanicMember:Create(name)
  table.insert(self.members, member)
  AceEvent:SendMessage("MEMBER_UPDATED", member.name)
  return member
end

function OrganizedPanicMembers:DeleteMember(member)
  for i = 1, #self.members do
    if self.members[i].name == member.name then
      self.members[i] = nil
      AceEvent:SendMessage("MEMBER_UPDATED", member.name)
      return
    end
  end
end

function OrganizedPanicMembers:InspectMember(member)
  member:SetInspecting()
  if member:IsPlayer() then
    local _, specialization, _, icon, role = GetSpecializationInfo(GetSpecialization())
    self:UpdateMemberSpecialization(member, specialization, icon)
  else
    if not CanInspect(member:GetGroupName()) then return end
    NS:debug("InspectMember", member.name, member.groupName, member.unitName, member.groupType)
    local unitName = member:GetUnitName()
    local success, error = pcall(OrganizedPanicMembers.NotifyInspect, self, unitName)
    if not success then
      NS:debug("InspectMember", "NotifyInspect failed", unitName, NS:ObjectToString(error))
    end
  end
end

function OrganizedPanicMembers:MemberInspected(member)
  member:SetInspected()
  AceEvent:SendMessage("MEMBER_INSPECTED", member.name, member.unitName)
  NS:debug("MemberInspected", member.name, member.groupName, GetTime() - member.lastInspectedTime)
end

function OrganizedPanicMembers:NotifyInspect(name)
  NS:debug("NotifyInspect", name)
  _G["NotifyInspect"](name)
end

function OrganizedPanicMembers:UpdateMemberSpecialization(member, specialization, icon)
  if not specialization then
    NS:debug("UpdateMemberSpecialization", "blank spec", member.name)
    return
  end
  member:SetSpecialization(specialization)
  member:SetSpecializationIcon(icon)
  member:Update()
  self:MemberInspected(member)
  AceEvent:SendMessage("MEMBER_UPDATED", member.name)
end

function OrganizedPanicMembers:FindMemberByName(name)
  for i, member in pairs(self:GetMembers()) do
    if member:GetName() == name then return member end
  end
end

function OrganizedPanicMembers:FindMemberByUnitId(unitId)
  for i, member in pairs(self:GetMembers()) do
    if member:GetGUID() == unitId then return member end
    if unitId == "player" and member:IsPlayer() then return member end
  end
end

function OrganizedPanicMembers:FindMemberByGroupName(groupName)
  for i, member in pairs(self:GetMembers()) do
    if member:GetGroupName() == groupName then return member end
    if groupName == "player" and member:IsPlayer() then return member end
  end
end

function OrganizedPanicMembers:FindMemberByUnitName(unitName)
  return self:FindMemberByName(OrganizedPanicCommon:GetUnitNameWithRealm(unitName))
end

function OrganizedPanicMembers:FindPlayer()
  for i, member in pairs(self:GetMembers()) do
    if member:IsPlayer() then return member end
  end
end

function OrganizedPanicMembers:GetMembers()
  return self.members or {}
end
